import numpy as np
import torch
import torch.nn as nn

__all__ = ['Kernel', 'RBFKernel', 'PeriodicKernel']


class Kernel(nn.Module):
    """A base class for GP kernels.
    """
    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, x1, x2, diag=False):
        raise NotImplementedError


class RBFKernel(Kernel):
    """A radial basis function (squared exponential) kernel.

    :param lengthscale (float, optional): initial lengthscale(s).
    :param scale (float, optional): initial scale.
    """
    def __init__(self, lengthscale=1., scale=1., **kwargs):
        super().__init__(**kwargs)

        self.lengthscale = nn.Parameter(torch.tensor(lengthscale),
                                        requires_grad=True)
        self.scale = nn.Parameter(torch.tensor(scale), requires_grad=True)

    def forward(self, x1, x2, diag=False):
        """Returns the covariance matrix defined by the RBF kernel."""
        x1 = x1.unsqueeze(1) if len(x1.shape) == 1 else x1
        x2 = x2.unsqueeze(1) if len(x2.shape) == 1 else x2

        assert x1.shape[-1] == x2.shape[-1], 'Inputs are different dimensions.'

        if not diag:
            x1 = x1.unsqueeze(1)
            x2 = x2.unsqueeze(0)
        else:
            assert x1.shape == x2.shape, 'Inputs must be the same shape.'

        # [M1, M2, D] or [M, D] if diag.
        sd = (x1 - x2) ** 2
        sd.clamp_min_(0)

        # Apply lengthscale and sum over dimensions.
        sd_ = (sd / self.lengthscale ** 2).sum(-1)
        cov = self.scale ** 2 * (-sd_).exp()

        return cov


class PeriodicKernel(Kernel):
    """A periodic kernel.

    :param lengthscale (float): initial lengthscale(s).
    :param period (float): initial period(s).
    :param scale (float): initial scale.
    """
    def __init__(self, lengthscale=1., period=1., scale=1., **kwargs):
        super().__init__(**kwargs)

        self.lengthscale = nn.Parameter(torch.tensor(lengthscale),
                                        requires_grad=True)
        self.scale = nn.Parameter(torch.tensor(scale), requires_grad=True)
        self.raw_period = nn.Parameter(torch.tensor(period).log(),
                                       requires_grad=True)

    def forward(self, x1, x2, diag=False):
        """ Returns the covariance matrix defined by the periodic kernel."""
        x1 = x1.unsqueeze(1) if len(x1.shape) == 1 else x1
        x2 = x2.unsqueeze(1) if len(x2.shape) == 1 else x2

        assert x1.shape[-1] == x2.shape[-1], 'Inputs are different dimensions.'

        if not diag:
            x1 = x1.unsqueeze(1)
            x2 = x2.unsqueeze(0)
        else:
            assert x1.shape == x2.shape, 'Inputs must be the same shape.'

        # [M1, M2, D] or [M, D] if diag.
        ad = (x1 - x2).abs()

        # Apply period.
        ad_ = 2 * (np.pi * ad / self.raw_period.exp()).sin() ** 2

        # Apply lengthscale and sum over dimensions.
        ad_ = (ad_ / self.lengthscale ** 2).sum(-1)
        cov = self.scale ** 2 * (-ad_).exp()

        return cov



import torch.nn as nn

__all__ = ['Likelihood']


class Likelihood(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, z):
        raise NotImplementedError

    def log_prob(self, z, x):
        px = self(z)

        return px.log_prob(x)



import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.distributions import Bernoulli
from .networks import LinearNN
from .base import Likelihood

__all__ = ['Bernoulli', 'NNBernoulli']


class Bernoulli(Likelihood):

    def __init__(self):
        super().__init__()

        self.loglikelihood = nn.BCEWithLogitsLoss(reduction='none')

    def forward(self, z):
        px = Bernoulli(torch.sigmoid(z))

        return px

    def log_prob(self, z, x):
        return self.loglikelihood(z, x)


class NNBernoulli(nn.Module):
    """A fully connected neural network for parameterising a Bernoulli
    distribution.
    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, hidden_dims=(64, 64),
                 nonlinearity=F.relu):
        super().__init__()

        self.network = LinearNN(in_dim, out_dim, hidden_dims, nonlinearity)
        self.likelihood = Bernoulli()

    def forward(self, z, x):
        mu = self.network(z)

        return self.likelihood(mu, x)

    def predict(self, z):
        mu = self.network(z)

        return self.likelihood.log_prob(mu)



import copy
import torch
import torch.nn as nn

from torch.distributions import MultivariateNormal, Normal, kl_divergence
from sgpvae.kernels import KernelList
from sgpvae.utils.matrix import add_diagonal

from .base import VAE

__all__ = ['GPVAE', 'SGPVAE']

JITTER = 1e-5


class GPVAE(VAE):
    """The GP-VAE model.

    :param likelihood: likelihood function, p(x|f).
    :param variational_dist: variational distribution, l(f|x).
    :param latent_dim (int): the dimension of latent space.
    :param kernel: GP kernel.
    :param add_jitter (bool, optional): whether to add jitter to the GP prior
    covariance matrix.
    """
    def __init__(self, likelihood, variational_dist, latent_dim, kernel,
                 add_jitter=False):
        super().__init__(likelihood, variational_dist, latent_dim)

        self.add_jitter = add_jitter

        if not isinstance(kernel, list):
            kernels = [copy.deepcopy(kernel) for _ in range(latent_dim)]
            self.kernels = KernelList(kernels)

        else:
            assert len(kernel) == latent_dim, 'Number of kernels must be ' \
                                              'equal to the latent dimension.'
            self.kernels = KernelList(copy.deepcopy(kernel))

    def pf(self, x, diag=False):
        """Return latent prior."""
        pf_mu = torch.zeros(self.latent_dim, x.shape[0])
        pf_cov = self.kernels.forward(x, x, diag)

        if self.add_jitter:
            # Add jitter to improve condition number.
            pf_cov = add_diagonal(pf_cov, JITTER)

        pf_chol = pf_cov.cholesky()
        pf = MultivariateNormal(pf_mu, scale_tril=pf_chol)

        return pf

    def lf(self, y, mask=None):
        """Return the approximate likelihood."""
        lf = self.variational_dist(y, mask)

        return lf

    def qf(self, x=None, y=None, pf=None, lf=None, mask=None, diag=False,
           x_test=None):
        if pf is None:
            pf = self.pf(x, diag)

        # Mean, covariance and Cholesky factor.
        pf_cov, pf_chol = pf.covariance_matrix, pf.scale_tril

        if lf is None:
            lf = self.lf(y, mask)

        # Reshape.
        lf_mu = lf.mean.transpose(0, 1)
        lf_sigma = lf.stddev.transpose(0, 1)
        lf_cov = lf_sigma.pow(2).diag_embed()

        # A = I + Lf^{-1} \Sigma_{\phi} Lf^{-T}
        a = torch.triangular_solve(lf_cov.pow(0.5), pf_chol, upper=False)[0]
        a = a.matmul(a.transpose(-1, -2))
        a = add_diagonal(a, 1)
        a_chol = torch.cholesky(a)

        # B = La^{-1} * Lf^{-1} * \mu_{\phi}.
        b = torch.triangular_solve(lf_mu.unsqueeze(2), pf_chol, upper=False)[0]
        b = torch.triangular_solve(b, a_chol, upper=False)[0]

        if x_test is not None:
            # GP prior.
            ps = self.pf(x_test)
            ps_cov = ps.covariance_matrix

            # GP conditional prior.
            ksf = self.kernels.forward(x_test, x)
            kfs = ksf.transpose(-1, -2)

            # C = Lf^{-1} * Kfs.
            c = torch.triangular_solve(kfs, pf_chol, upper=False)[0]

            # D = La^{-1} * Lf^{-1} * Kfs
            d = torch.triangular_solve(c, a_chol, upper=False)[0]

            qs_mu = torch.triangular_solve(b, a_chol.transpose(-1, -2),
                                           upper=True)[0]
            qs_mu = c.transpose(-1, -2).matmul(qs_mu).squeeze(2)
            qs_cov = ps_cov - d.transpose(-1, -2).matmul(d)
            qs = MultivariateNormal(qs_mu, qs_cov)

            return qs

        else:
            # C = La^{-1} * Lf.
            c = torch.triangular_solve(pf_chol.transpose(-1, -2), a_chol,
                                       upper=False)[0]

            qf_mu = c.transpose(-1, -2).matmul(b).squeeze(2)
            qf_cov = pf_cov - c.transpose(-1, -2).matmul(c)
            qf = MultivariateNormal(qf_mu, qf_cov)

            return qf

    def elbo(self, x, y, mask=None, mask_q=None, num_samples=1):
        """Monte Carlo estimate of the evidence lower bound."""
        # Use y_q and mask_q to obtain approximate likelihoods.
        if mask_q is None:
            mask_q = mask

        pf = self.pf(x)
        lf = self.lf(y, mask_q)
        qf = self.qf(pf=pf, lf=lf)

        # KL(q(f) || p(f)) term.
        kl = kl_divergence(qf, pf).sum()

        # log p(y|f) term.
        qf_sigma = torch.stack(
            [cov.diag().pow(0.5) for cov in qf.covariance_matrix])
        qf_marginals = Normal(qf.mean, qf_sigma)
        f_samples = qf_marginals.rsample((num_samples,))

        log_py_f = 0
        for f in f_samples:
            if mask is not None:
                log_py_f += (self.likelihood.log_prob(f.T, y) * mask).sum()
            else:
                log_py_f += self.likelihood.log_prob(f.T, y).sum()

        log_py_f /= num_samples
        elbo = log_py_f - kl

        return elbo / y.shape[0]

    def predict_y(self, x, y=None, mask=None, x_test=None, num_samples=1):
        """Sample predictive posterior."""
        if y is None:
            if x_test is None:
                qf = self.pf(x)
            else:
                qf = self.pf(x_test)
        else:
            qf = self.qf(x=x, y=y, mask=mask, x_test=x_test)

        f_samples = qf.sample((num_samples,))

        y_mus, y_sigmas, y_samples = [], [], []
        for f in f_samples:
            # Output conditional posterior distribution.
            py_f = self.likelihood(f.T)
            y_mus.append(py_f.mean)
            y_samples.append(py_f.sample())

        y_mu = torch.stack(y_mus).mean(0).detach()
        y_sigma = torch.stack(y_samples).std(0).detach()

        return y_mu, y_sigma, y_samples


class SGPVAE(GPVAE):
    """The SGP-VAE model.

    :param likelihood: likelihood function, p(x|f).
    :param variational_dist: variational distribution, l(f|x).
    :param latent_dim (int): the dimension of latent space.
    :param kernel: GP kernel.
    :param z (tensor): initial inducing point locations.
    :param add_jitter (bool, optional): whether to add jitter to the GP
    prior covariance matrix.
    :param fixed_inducing (bool, optional): whether to fix the inducing
    points.
    """
    def __init__(self, likelihood, variational_dist, latent_dim, kernel, z,
                 add_jitter=False, fixed_inducing=False):
        super().__init__(likelihood, variational_dist, latent_dim, kernel,
                         add_jitter)

        if fixed_inducing:
            self.z = nn.Parameter(z, requires_grad=False)
        else:
            self.z = nn.Parameter(z, requires_grad=True)

    def qf(self, x, y=None, pu=None, lf=None, mask=None, diag=False,
           x_test=None, full_cov=False):
        if pu is None:
            pu = self.pf(self.z)

        # Cholesky factor.
        pu_chol = pu.scale_tril

        if lf is None:
            lf = self.lf(y, mask)

        # Reshape.
        lf_mu = lf.mean.transpose(0, 1)
        lf_sigma = lf.stddev.transpose(0, 1)
        lf_precision = lf_sigma.pow(-2).diag_embed()

        # GP conditional prior.
        kfu = self.kernels.forward(x, self.z)
        kuf = kfu.transpose(-1, -2)

        # A = Lu^{-1} * Kuf.
        a = torch.triangular_solve(kuf, pu_chol, upper=False)[0]

        # B = I + A * \Sigma_{\phi}^{-1} * A^T.
        b = a.matmul(lf_precision).matmul(a.transpose(-1, -2))
        b = add_diagonal(b, 1)
        b_chol = torch.cholesky(b)

        # c = Lb^{-1} * Lu^{-1} * Kuf * \Sigma_{\phi}^{-1} * \mu_{\phi}
        c = a.matmul(lf_precision).matmul(lf_mu.unsqueeze(2))
        c = torch.triangular_solve(c, b_chol, upper=False)[0]

        # d = Lu^{-T} * Lb^{-T} * c
        d = torch.triangular_solve(c, b_chol.transpose(-1, -2), upper=True)[0]
        d = torch.triangular_solve(d, pu_chol.transpose(-1, -2), upper=True)[0]

        if x_test is not None:
            # GP prior.
            ps = self.pf(x_test, diag=(not full_cov))
            ps_cov = ps.covariance_matrix

            # GP conditional prior.
            ksu = self.kernels.forward(x_test, self.z)
            kus = ksu.transpose(-1, -2)

            # e = Lu^{-1} * Kus.
            e = torch.triangular_solve(kus, pu_chol, upper=False)[0]

            # g = Lb^{-1} * e
            g = torch.triangular_solve(e, b_chol, upper=False)[0]

            qs_mu = ksu.matmul(d).squeeze(2)
            qs_cov = (ps_cov - e.transpose(-1, -2).matmul(e)
                      + g.transpose(-1, -2).matmul(g))
            qs = MultivariateNormal(qs_mu, qs_cov)

            return qs, None

        else:
            pf = self.pf(x, diag)
            # Covariance and Cholesky factor.
            pf_cov, pf_chol = pf.covariance_matrix, pf.scale_tril

            # e = Lb^{-1} * a
            e = torch.triangular_solve(a, b_chol, upper=False)[0]

            qf_cov = (pf_cov - a.transpose(-1, -2).matmul(a)
                      + e.transpose(-1, -2).matmul(e))
            qf_mu = kfu.matmul(d).squeeze(2)
            qf = MultivariateNormal(qf_mu, qf_cov)

            # g = Lb^{-1} * Lu.
            g = torch.triangular_solve(pu_chol, b_chol, upper=False)[0]

            qu_mu = torch.triangular_solve(c, b_chol.transpose(-1, -2),
                                           upper=True)[0]
            qu_mu = pu_chol.matmul(qu_mu).squeeze(2)
            qu_cov = g.transpose(-1, -2).matmul(g)
            qu = MultivariateNormal(qu_mu, qu_cov)

            return qf, qu

    def elbo(self, x, y, mask=None, num_samples=1):
        """Monte Carlo estimate of the evidence lower bound."""
        pu = self.pf(self.z)
        lf = self.lf(y, mask)
        qf, qu = self.qf(x, pu=pu, lf=lf)

        # KL(q(u) || p(u)) term.
        kl = kl_divergence(qu, pu).sum()

        # log p(y|f) term.
        qf_sigma = torch.stack(
            [cov.diag().pow(0.5) for cov in qf.covariance_matrix])
        qf_marginals = Normal(qf.mean, qf_sigma)
        f_samples = qf_marginals.rsample((num_samples,))
        log_py_f = 0
        for f in f_samples:
            if mask is not None:
                log_py_f += (self.likelihood.log_prob(f.T, y) * mask).sum()
            else:
                log_py_f += self.likelihood.log_prob(f.T, y).sum()

        log_py_f /= num_samples
        elbo = log_py_f - kl

        return elbo / y.shape[0]

    def predict_y(self, x, y=None, mask=None, x_test=None, full_cov=True,
                  num_samples=1):
        """Sample predictive posterior."""
        if y is None:
            if x_test is None:
                qf = self.pf(x)
            else:
                qf = self.pf(x_test)
        else:
            qf = self.qf(x=x, y=y, mask=mask, x_test=x_test,
                         full_cov=full_cov)[0]

        if full_cov:
            f_samples = qf.sample((num_samples,))
        else:
            qf_sigma = torch.stack(
                [cov.diag().pow(0.5) for cov in qf.covariance_matrix])
            qf_marginals = Normal(qf.mean, qf_sigma)
            f_samples = qf_marginals.sample((num_samples,))

        y_mus, y_sigmas, y_samples = [], [], []
        for f in f_samples:
            # Output conditional posterior distribution.
            py_f = self.likelihood(f.T)
            y_mus.append(py_f.mean)
            y_samples.append(py_f.sample())

        y_mu = torch.stack(y_mus).mean(0).detach()
        y_sigma = torch.stack(y_samples).std(0).detach()

        return y_mu, y_sigma, y_samples



import torch.nn.functional as F

from .networks import LinearNN
from .base import Likelihood
from .gaussian import HomoGaussian

__all__ = ['GPRNHomoGaussian', 'GPRNNNHomoGaussian', 'NNGPRNHomoGaussian']


class GPRNHomoGaussian(Likelihood):
    """The likelihood function described in Gaussian Process Regresssion
    Networks with homoscedastic noise.

    :param f_dim (int): dimension of the latent space.
    :param out_dim (int): dimension of the output variable.
    :param sigma (float, optional): if not None, sets the initial
    homoscedastic output sigma.
    :param sigma_grad (bool, optional): whether to train the homoscedastic
    output sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layeres.
    """
    def __init__(self, f_dim, out_dim, sigma=None, sigma_grad=True,
                 min_sigma=1e-3, w_transform=None):
        super().__init__()

        self.f_dim = f_dim
        self.w_dim = out_dim
        self.w_transform = w_transform
        self.likelihood = HomoGaussian(out_dim, sigma, sigma_grad, min_sigma)

    def forward(self, z):
        f, w = z[:, :self.f_dim], z[:, self.f_dim:]
        w = w.reshape(-1, self.w_dim, self.f_dim)

        # Transform weights of GPRN if applicable.
        if self.w_transform is not None:
            w = self.w_transform(w)

        mu = w.matmul(f.unsqueeze(-1)).squeeze(-1)

        return self.likelihood(mu)


class GPRNNNHomoGaussian(Likelihood):
    """The likelihood function described in Gaussian Process Regresssion
    Networks with a fully connected neural network stuck on top with
    homoscedastic noise.

    :param f_dim (int): dimension of the latent space.
    :param w_dim (int): dimension of the first hidden layer.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param sigma (float, optional): if not None, sets the initial
    homoscedastic output sigma.
    :param sigma_grad (bool, optional): whether to train the homoscedastic
    output sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param w_transform (function, optional): transform to apply to the GPRN
    weights.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """

    def __init__(self, f_dim, w_dim, out_dim, hidden_dims=(64, 64),
                 sigma=None, sigma_grad=True, min_sigma=1e-3, w_transform=None,
                 nonlinearity=F.relu):
        super().__init__()

        self.f_dim = f_dim
        self.w_dim = w_dim
        self.w_transform = w_transform
        self.network = LinearNN(w_dim, out_dim, hidden_dims, nonlinearity)
        self.likelihood = HomoGaussian(out_dim, sigma, sigma_grad, min_sigma)

    def forward(self, z):
        f, w = z[:, :self.f_dim], z[:, self.f_dim:]
        w = w.reshape(-1, self.w_dim, self.f_dim)

        # Transform weights of GPRN if applicable.
        if self.w_transform is not None:
            w = self.w_transform(w)
        
        mu = self.network(w.matmul(f.unsqueeze(-1)).squeeze(-1))

        return self.likelihood(mu)


class NNGPRNHomoGaussian(Likelihood):
    """Fully connected neural network followed by the GPRN with
    homoscedastic noise.

    :param f_dim (int): dimension of the latent space.
    :param w_dim (int): dimension of the final hidden layer.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param sigma (float, optional): if not None, sets the initial
    homoscedastic output sigma.
    :param sigma_grad (bool, optional): whether to train the homoscedastic
    output sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layeres.
    """

    def __init__(self, f_dim, w_dim, out_dim, hidden_dims=(64, 64),
                 sigma=None, sigma_grad=True, min_sigma=1e-3,
                 nonlinearity=F.relu):
        super().__init__()

        self.f_dim = f_dim
        self.w_dim = w_dim
        self.out_dim = out_dim
        self.network = LinearNN(f_dim, w_dim, hidden_dims, nonlinearity)
        self.likelihood = HomoGaussian(out_dim, sigma, sigma_grad, min_sigma)

    def forward(self, z):
        f, w = z[:, :self.f_dim], z[:, self.f_dim:]
        f = self.network(f)
        w = w.reshape(-1, self.out_dim, self.w_dim)
        mu = w.matmul(f.unsqueeze(-1)).squeeze(-1)

        return self.likelihood(mu)



import os
import pickle
import pandas as pd
import numpy as np

from .data import data_path, resource

__all__ = ['load']


# Keep fixed for reproducibility.
np.random.seed(1)
cache_weekly = data_path('japan', 'weekly', 'experiment.pickle')


def load():
    _fetch()

    # If path does not exists, generate DataFrames.
    if not os.path.exists(cache_weekly):
        _parse()

    with open(cache_weekly, 'rb') as f:
        return pickle.load(f)


def _fetch():
    # All data.
    resource(
        target=data_path('japan', '80s.csv'),
        url='https://www.dropbox.com/s/kokgr6ekb0gh7wq/japan_80s.csv?dl=1')


def _parse():
    df = pd.read_csv(data_path('japan', '80s.csv'))

    # Extract 1980 and 1981 data.
    df = df[(df['date'].str[:4] == '1980') | (df['date'].str[:4] == '1981')]

    # Rescale TAVG.
    df['TAVG'] = df['TAVG'] / 10

    # Convert dates to Datetime.
    df['date'] = pd.DatetimeIndex(df['date'])

    # Create weekly groups.
    n = 7
    df['week'] = (df['date'] - df.iloc[0]['date']).dt.days // n
    df['day'] = (df['date'] - df.iloc[0]['date']).dt.days % n

    observations = ['TMAX', 'TMIN', 'TAVG']
    test_idx = [[] for _ in observations]
    for week, week_df in df.groupby('week'):
        # Set TAVG for middle three days to missing.
        mid_df = week_df[week_df['day'].isin([1, 2, 3, 4, 5])]
        test_idx[2] += list(mid_df.index)

        # Set 25% of TMAX and TMIN values to missing.
        m = len(week_df) // 4
        test_idx[0] += list(np.random.choice(list(week_df.index), m,
                                             replace=False))
        test_idx[1] += list(np.random.choice(list(week_df.index), m,
                                             replace=False))

    # Train indices are all those that aren't in test indices.
    train_idx = [list(set(df.index.tolist()) - set(idx)) for idx in test_idx]

    train = df.copy()
    test = df.copy()

    for obs, train_idx_, test_idx_ in zip(observations, train_idx, test_idx):
        train[obs][test_idx_] = np.nan
        test[obs][train_idx_] = np.nan

    # Set all other nans in test df.
    for obs in ['PRCP', 'SNWD']:
        test[obs][:] = np.nan

    train_1980 = train[train['date'].dt.year == 1980]
    train_1981 = train[train['date'].dt.year == 1981]
    test_1980 = test[test['date'].dt.year == 1980]
    test_1981 = test[test['date'].dt.year == 1981]

    # Train on all data from 1980.
    train = df[df['date'].dt.year == 1980]

    data = {'all': df}
    names = ['train', 'train_1980', 'train_1981', 'test_1980', 'test_1981']
    dfs = [train, train_1980, train_1981, test_1980, test_1981]
    # Extract stations per group.
    for name, df in zip(names, dfs):
        per_week = []
        for week, week_df in df.groupby('week'):
            per_week.append(week_df.copy())

        data[name] = per_week

    # Extract dates.
    data['dates_1980'] = [x.iloc[0].week for x in data['train_1980']]
    data['dates_1981'] = [x.iloc[0].week for x in data['train_1981']]

    # Save experiment data.
    if not os.path.exists(os.path.dirname(cache_weekly)):
        os.makedirs(os.path.dirname(cache_weekly))

    with open(cache_weekly, 'wb') as f:
        pickle.dump(data, f)



import copy
import torch

from torch.utils.data import Dataset

__all__ = ['TupleDataset', 'MetaTupleDataset']


class TupleDataset(Dataset):
    """An iterable over a dataset of tuples (x, y).

    :param x: (tensor) input data.
    :param y: (tensor) output data.
    :param missing: (bool, optional) whether data contains missing data.
    """
    def __init__(self, x, y, missing=False):
        super().__init__()

        assert len(x) == len(y), 'x and y must be the same length.'

        if len(x.shape) == 1:
            # Ensure inputs are 2-dimensional.
            self.x = x.unsqueeze(1)
        else:
            self.x = x

        if missing:
            self.y = copy.deepcopy(y)
            self.m = torch.ones_like(y).fill_(True)

            # Identify nan values and replace with 0.
            m_idx = torch.isnan(y)
            self.m[m_idx] = False
            self.y[m_idx] = 0.
        else:
            self.y = y

        self.missing = missing

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        x = self.x[idx]
        y = self.y[idx]

        if self.missing:
            m = self.m[idx]
            return x, y, m, idx
        else:
            return x, y, idx

    def dataset(self):
        idx = list(range(len(self)))

        return self.__getitem__(idx)


class MetaTupleDataset(Dataset):
    """An iterable over datasets of tuples (x, y).

   :param x: (list) list of input data.
   :param y: (list) list of output data.
   :param missing: (bool, optional) whether data contains missing data.
   """
    def __init__(self, x, y, missing=False):
        super().__init__()

        assert len(x) == len(y), 'x and y must be the same length.'

        self.x = x

        if missing:
            self.y = copy.deepcopy(y)

            self.m = list(map(lambda y_: torch.ones_like(y_).fill_(True), y))

            # Identify nan values and replace with 0.
            m_idx = list(map(lambda y_: torch.isnan(y_), y))

            for i, idx in enumerate(m_idx):
                self.m[i][idx] = False
                self.y[i][idx] = 0.
        else:
            self.y = y

        self.missing = missing

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        x = self.x[idx]
        y = self.y[idx]

        if self.missing:
            m = self.m[idx]
            return x, y, m, idx
        else:
            return x, y, idx

    def dataset(self):
        idx = list(range(len(self)))

        return self.__getitem__(idx)



import datetime
import os
import shutil
import subprocess

import numpy as np
import pandas as pd
import requests
import urllib.request
from contextlib import closing

__all__ = ['DependencyError',
           'resource',
           'dependency',
           'asserted_dependency',
           'split_df',
           'data_path',
           'date_to_decimal_year']


class DependencyError(AssertionError):
    """Exception raised in case of an erroneous dependency."""


def resource(target, url, post=False, **kw_args):
    """Specify a dependency on an online resource.

    Further takes in keyword arguments that are passed to the appropriate method
    from :mod:`requests` or :mod:`urllib`.

    Args:
        target (str): Target file.
        url (str): Source URL.
        post (bool, optional): Make a POST request instead of a GET request.
            Only applicable if the URL starts with "http" or "https". Defaults
            to `False`.
    """
    if not os.path.exists(target):
        # Ensure that all directories in the path exist.
        make_dirs(target)

        # If the URL starts with "ftp", use the :mod:`urllib` library.
        # Otherwise, use the :mod:`requests` library.
        if url.startswith('ftp'):
            with closing(urllib.request.urlopen(url, **kw_args)) as r:
                with open(target, 'wb') as f:
                    shutil.copyfileobj(r, f)
        else:
            request = requests.post if post else requests.get
            with request(url, stream=True, **kw_args) as r:
                with open(target, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)


def dependency(target, source, commands):
    """Specify a dependency that is generated from an existing file.

    Args:
        target (str): Target file.
        source (str): Source file.
        commands (list[str]): List of commands to generate target file.
    """
    if not os.path.exists(target):
        # Check that the source exists.
        if not os.path.exists(source):
            raise DependencyError(f'Source "{source}" asserted to exist, '
                                  f'but it does not.')

        # Save current working directory.
        current_wd = os.getcwd()

        # Ensure that all directories in the path exist.
        make_dirs(target)

        # Perform commands.
        for command in commands:
            # Change working directory to directory of target file, run
            # command, and restore working directory afterwards.
            os.chdir(os.path.dirname(target))
            subprocess.call(command, shell=True)
            os.chdir(current_wd)


def asserted_dependency(target):
    """Specify a dependency that cannot be fetched.

    Args:
        target (str): Target file.
    """
    if not os.path.exists(target):
        raise DependencyError(f'Dependency "{target}" is asserted to exist, '
                              f'but it does not, and it cannot be '
                              f'automatically fetched. Please put the file '
                              f'into place manually.')


def make_dirs(path):
    """Make the directories in the path of a file.

    Args:
        path (url): Path of a file.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)


def data_path(*xs):
    """Get the path of a data file.

    Args:
        *xs (str): Parts of the path.

    Returns:
        str: Absolute path.
    """
    return os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        os.pardir,
                                        os.pardir,
                                        'data',
                                        *xs))


def split_df(df, index_range, columns, iloc=False):
    """Split a data frame by selecting from columns a particular range.

    Args:
        df (:class:`pd.DataFrame`): Data frame to split.
        index_range (tuple): Tuple containing lower and upper limit of the
            range to split the index by. If `index_range = (a, b)`, then
            `[a, b)` is taken.
        columns (list[object]): Columns to select.
        iloc (bool, optional): The index range is the integer location instead
            of the index value. Defaults to `False`.

    Returns:
        tuple[:class:`pd.DataFrame`]: Selected rows from selected columns
            and the remainder.
    """
    if iloc:
        inds = np.arange(df.shape[0])
        rows = (inds >= index_range[0]) & (inds < index_range[1])
    else:
        rows = (df.index >= index_range[0]) & (df.index < index_range[1])
    selected = pd.DataFrame([df[name][rows] for name in columns]).T
    remainder = pd.DataFrame([df[name][~rows] for name in columns] +
                             [df[name] for name in
                              set(df.columns) - set(columns)]).T

    # Fix order of columns.
    selected_inds = [i for i, c in enumerate(df.columns) if c in columns]
    selected = selected.reindex(df.columns[np.array(selected_inds)], axis=1)
    remainder = remainder.reindex(df.columns, axis=1)

    return selected, remainder


def date_to_decimal_year(date, format):
    """Convert a date to decimal year.

    Args:
        date (str): Date as a string.
        format (str): Format of the dat.

    Returns:
        float: Decimal year corresponding to the date.
    """
    date = datetime.datetime.strptime(date, format)
    start = datetime.date(date.year, 1, 1).toordinal()
    year_length = datetime.date(date.year + 1, 1, 1).toordinal() - start
    return date.year + float(date.toordinal() - start) / year_length



import argparse
import os
import pickle

__all__ = ['str2bool', 'save']


def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def save(args, metrics):
    if 'results_dir' in args.keys():
        results_dir = args['results_dir'] + args['model']
    else:
        results_dir = '_results/' + args['model']

    if os.path.isdir(results_dir):
        i = 1
        while os.path.isdir(results_dir + '_' + str(i)):
            i += 1

        results_dir += '_' + str(i)

    os.makedirs(results_dir, exist_ok=True)

    # Pickle args and metrics.
    with open(results_dir + '/args.pkl', 'wb') as f:
        pickle.dump(args, f)

    with open(results_dir + '/metrics.pkl', 'wb') as f:
        pickle.dump(metrics, f)

    # Save args and results in text format.
    with open(results_dir + '/results.txt', 'w') as f:
        f.write('Args: \n')
        if isinstance(args, list):
            for d in args:
                f.write(str(d) + '\n')
        else:
            f.write(str(args) + '\n')

        f.write('\nMetrics: \n')
        for (key, value) in metrics.items():
            f.write('{}: {}\n'.format(key, value))



import torch
import torch.nn as nn

from torch.distributions import Normal, kl_divergence

__all__ = ['VAE']


class VAE(nn.Module):
    """VAE with standard normal prior.

    :param likelihood: likelihood function, p(x|f).
    :param variational_dist: variational distribution, q(f|x).
    :param latent_dim (int): latent space dimensionality.
    """
    def __init__(self, likelihood, variational_dist, latent_dim):
        super().__init__()

        self.likelihood = likelihood
        self.variational_dist = variational_dist
        self.latent_dim = latent_dim

    def pf(self):
        """Return latent prior."""
        pf_mu = torch.zeros(self.latent_dim)
        pf_sigma = torch.ones(self.latent_dim)
        pf = Normal(pf_mu, pf_sigma)

        return pf

    def qf(self, y, mask=None):
        """Return latent approximate posterior."""
        qf = self.variational_dist(y, mask)

        return qf

    def elbo(self, y, mask=None, num_samples=1):
        """Monte Carlo estimate of the evidence lower bound."""
        pf = self.pf()
        qf = self.qf(y, mask)

        # KL(q(f) || p(f) term.
        kl = kl_divergence(qf, pf).sum()

        # log p(y|f) term.
        f_samples = qf.rsample((num_samples,))
        log_py_f = 0
        for f in f_samples:
            log_py_f += (self.likelihood.log_prob(f, y) * mask).sum()

        log_py_f /= num_samples
        elbo = log_py_f - kl

        return elbo / y.shape[0]

    def predict_y(self, y=None, num_samples=1):
        """Sample predictive posterior."""
        if y is None:
            qf = self.pf()
        else:
            qf = self.qf(y)

        f_samples = qf.sample((num_samples,))

        y_mus, y_sigmas, y_samples = [], [], []
        for f in f_samples:
            # Output conditional posterior distribution.
            py_f = self.likelihood(f)
            y_mus.append(py_f.mean)
            y_samples.append(py_f.sample())

        y_mu = torch.stack(y_mus).mean(0).detach()
        y_sigma = torch.stack(y_samples).std(0).detach()

        return y_mu, y_sigma, y_samples



import torch
import torch.nn as nn

from torch.nn import functional as F
from torch.distributions import Normal
from .base import Likelihood
from .gaussian import NNHeteroGaussian
from .networks import LinearNN

__all__ = ['IndexNet', 'FactorNet', 'PointNet']


class FactorNet(Likelihood):
    """FactorNet from Sparse Gaussian Process Variational Autoencoders.

    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param h_dims (list, optional): dimensions of hidden layers.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param initi_sigma (float, optional): sets the initial output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, h_dims=(64, 64), min_sigma=1e-3,
                 init_sigma=None, nonlinearity=F.relu):
        super().__init__()

        self.out_dim = out_dim

        # Rescale sigmas for multiple outputs.
        if init_sigma is not None:
            init_sigma = init_sigma * in_dim ** 0.5

        min_sigma = min_sigma * in_dim ** 0.5

        # A network for each dimension.
        self.networks = nn.ModuleList()
        for _ in range(in_dim):
            self.networks.append(NNHeteroGaussian(
                1, out_dim, h_dims, min_sigma, init_sigma,
                nonlinearity=nonlinearity))

    def forward(self, z, mask=None):
        """Returns parameters of a diagonal Gaussian distribution."""
        np_1 = torch.zeros(z.shape[0], z.shape[1], self.out_dim)
        np_2 = torch.zeros_like(np_1)

        # Pass through individual networks.
        for dim, z_dim in enumerate(z.transpose(0, 1)):
            if mask is not None:
                idx = torch.where(mask[:, dim])[0]
                z_in = z_dim[idx].unsqueeze(1)

                # Don't pass through if no inputs.
                if len(z_in) != 0:
                    pz = self.networks[dim](z_in)
                    mu, sigma = pz.mean, pz.stddev
                    np_1[idx, dim, :] = mu / sigma ** 2
                    np_2[idx, dim, :] = - 1. / (2. * sigma ** 2)
            else:
                z_in = z_dim.unsqueeze(1)
                pz = self.networks[dim](z_in)
                mu, sigma = pz.mean, pz.stddev
                np_1[:, dim, :] = mu / sigma ** 2
                np_2[:, dim, :] = -1. / (2. * sigma ** 2)

        # Sum natural parameters.
        np_1 = torch.sum(np_1, 1)
        np_2 = torch.sum(np_2, 1)
        sigma = (- 1. / (2. * np_2)) ** 0.5
        mu = np_1 * sigma ** 2.

        pz = Normal(mu, sigma)

        return pz


class IndexNet(Likelihood):
    """IndexNet from Sparse Gaussian Process Variational Autoencoders.

    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param inter_dim (int): dimension of intermediate representation.
    :param h_dims (list, optional): dimension of the encoding function.
    :param rho_dims (list, optional): dimension of the shared function.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param init_sigma (float, optional): sets the initial output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, inter_dim, h_dims=(64, 64),
                 rho_dims=(64, 64), min_sigma=1e-3, init_sigma=None,
                 nonlinearity=F.relu):
        super().__init__()

        self.out_dim = out_dim
        self.inter_dim = inter_dim

        # A network for each dimension.
        self.networks = nn.ModuleList()
        for _ in range(in_dim):
            self.networks.append(LinearNN(1, inter_dim, h_dims, nonlinearity))

        # Takes the aggregation of the outputs from self.networks.
        self.rho = NNHeteroGaussian(
            inter_dim, out_dim, rho_dims, min_sigma, init_sigma, nonlinearity)

    def forward(self, z, mask=None):
        """Returns parameters of a diagonal Gaussian distribution."""
        out = torch.zeros(z.shape[0], z.shape[1], self.inter_dim)

        # Pass through individual networks.
        for dim, z_dim in enumerate(z.transpose(0, 1)):
            if mask is not None:
                idx = torch.where(mask[:, dim])[0]
                z_in = z_dim[idx].unsqueeze(1)

                # Don't pass through if no inputs.
                if len(z_in) != 0:
                    z_out = self.networks[dim](z_in)
                    out[idx, dim, :] = z_out

            else:
                z_in = z_dim.unsqueeze(1)
                z_out = self.networks[dim](z_in)
                out[:, dim, :] = z_out

        # Aggregation layer.
        out = torch.sum(out, 1)

        # Pass through shared network.
        pz = self.rho(out)

        return pz


class PointNet(Likelihood):
    """PointNet from Sparse Gaussian Process Variational Autoencoders.

    :param out_dim (int): dimension of the output variable.
    :param inter_dim (int): dimension of intermediate representation.
    :param h_dims (list, optional): dimension of the encoding function.
    :param rho_dims (list, optional): dimension of the shared function.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param init_sigma (float, optional): sets the initial output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, out_dim, inter_dim, h_dims=(64, 64), rho_dims=(64, 64),
                 min_sigma=1e-3, init_sigma=None, nonlinearity=F.relu):
        super().__init__()

        self.out_dim = out_dim
        self.inter_dim = inter_dim

        # Takes the index of the observation dimension and it's value.
        self.h = LinearNN(
            2, inter_dim, h_dims, nonlinearity)

        # Takes the aggregation of the outputs from self.h.
        self.rho = NNHeteroGaussian(
            inter_dim, out_dim, rho_dims, min_sigma, init_sigma, nonlinearity)

    def forward(self, z, mask=None):
        """Returns parameters of a diagonal Gaussian distribution."""
        out = torch.zeros(z.shape[0], z.shape[1], self.inter_dim)

        # Pass through first network.
        for dim, z_dim in enumerate(z.transpose(0, 1)):
            if mask is not None:
                idx = torch.where(mask[:, dim])[0]
                z_in = z_dim[idx].unsqueeze(1)
                z_in = torch.cat([z_in, torch.ones_like(z_in)*dim], 1)
                out[idx, dim, :] = self.h(z_in)
            else:
                z_in = z_dim.unsqueeze(1)
                torch.cat([z_in, torch.ones_like(z_in)*dim], 1)
                out[:, dim, :] = self.h(z_in)

        # Aggregation layer.
        out = torch.sum(out, 1)

        # Pass through second network.
        pz = self.rho(out)

        return pz



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from torch.distributions import Normal
from .networks import LinearNN
from .base import Likelihood

__all__ = ['HomoGaussian', 'AffineHomoGaussian', 'NNHomoGaussian',
           'NNHeteroGaussian']


class HomoGaussian(Likelihood):

    def __init__(self, dim, sigma=None, sigma_grad=True, min_sigma=1e-3):
        super().__init__()

        self.min_sigma = min_sigma

        if sigma is None:
            self.log_sigma = nn.Parameter(torch.zeros(dim),
                                          requires_grad=sigma_grad)
        else:
            self.log_sigma = nn.Parameter(torch.ones(dim) * np.log(sigma),
                                          requires_grad=sigma_grad)

    def forward(self, z):
        sigma = self.log_sigma.exp().clamp(min=self.min_sigma)
        px = Normal(z, sigma)

        return px


class AffineHomoGaussian(Likelihood):
    """Affine transformation of the input for parameterising d diagonal
    Gaussian distribution with homoscedastic noise.

    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param sigma (float, optional): initial homoscedastic output sigma.
    :param sigma_grad (float, optional): whether to train the homoscedastic
    ouput sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param init_weight (tensor, optional): initial weight.
    :param init_bias (teensor, optional): initial bias.
    """
    def __init__(self, in_dim, out_dim, sigma=None, sigma_grad=True,
                 min_sigma=1e-3, init_weight=None, init_bias=None):
        super().__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim

        if init_weight is None:
            init_weight = torch.ones(out_dim, in_dim) / in_dim
        else:
            init_weight = torch.tensor(init_weight)

        if init_bias is None:
            init_bias = torch.zeros(out_dim)
        else:
            init_bias = torch.tensor(init_bias)

        # Initial weight and bias of the affine transformation.
        self.weight = nn.Parameter(init_weight, requires_grad=True)
        self.bias = nn.Parameter(init_bias, requires_grad=True)

        self.likelihood = HomoGaussian(out_dim, sigma, sigma_grad, min_sigma)

    def forward(self, z):
        mu = self.weight.matmul(z.unsqueeze(2)).squeeze(2) + self.bias

        return self.likelihood(mu)


class NNHomoGaussian(Likelihood):
    """A fully connected neural network for parameterising a diagonal
    Gaussian distribution with homoscedastic noise.

    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param sigma (float, optional): if not None, sets the initial
    homoscedastic output sigma.
    :param sigma_grad (bool, optional): whether to train the homoscedastic
    output sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, hidden_dims=(64, 64), sigma=None,
                 sigma_grad=True, min_sigma=1e-3, nonlinearity=F.relu):
        super().__init__()

        self.network = LinearNN(in_dim, out_dim, hidden_dims, nonlinearity)
        self.likelihood = HomoGaussian(out_dim, sigma, sigma_grad, min_sigma)

    def forward(self, z):
        mu = self.network(z)

        return self.likelihood(mu)


class NNHeteroGaussian(Likelihood):
    """A fully connected neural network for parameterising a diagonal
    Gaussian distribution with heteroscedastic noise.
    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param sigma (float, optional): if not None, sets the initial
    homoscedastic output sigma.
    :param sigma_grad (bool, optional): whether to train the homoscedastic
    output sigma.
    :param min_sigma (float, optional): sets the minimum output sigma.
    :param init_sigma (float, optional): sets the initial output sigma.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, hidden_dims=(64, 64),
                 min_sigma=1e-3, init_sigma=None, nonlinearity=F.relu):
        super().__init__()

        self.out_dim = out_dim
        self.min_sigma = min_sigma
        self.network = LinearNN(in_dim, 2 * out_dim, hidden_dims, nonlinearity)

        if init_sigma is not None:
            self.network.layers[-1].bias.data[out_dim:] = torch.log(
                torch.exp(torch.tensor(init_sigma)) - 1)

    def forward(self, z, *args, **kwargs):
        output = self.network(z)
        mu = output[..., :self.out_dim]
        raw_sigma = output[..., self.out_dim:]
        sigma = F.softplus(raw_sigma) + self.min_sigma
        px = Normal(mu, sigma)

        return px



import torch

__all__ = ['add_diagonal']


def add_diagonal(x, val=1.):
    """Adds a value to the diagonal of a matrix.

    :param x: (tensor) matrix to modify.
    :param val: (float) value to add to the diagonal.
    """
    assert x.shape[-2] == x.shape[-1], 'x must be square.'

    d = (torch.ones(x.shape[-2]) * val).diag_embed()

    return x + d



import numpy as np

__all__ = ['mse', 'rmse', 'mll', 'mae', 'mre', 'smll', 'smse', 'srmse', 'r2']


def mre(mean, data):
    """Mean relative error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return ((mean - data).abs().sum(0) / data.abs().sum(0)).mean()


def mse(mean, data):
    """Mean squared error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return ((mean - data) ** 2).mean()


def smse(mean, data):
    """Standardised mean squared error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return mse(mean, data) / mse(data.mean(), data)


def rmse(mean, data):
    """Root mean squared error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return mse(mean, data) ** .5


def srmse(mean, data):
    """Standardised root mean squared error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return rmse(mean, data) / rmse(data.mean(), data)


def mae(mean, data):
    """Mean absolute error.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return np.abs(mean - data).mean()


def mll(mean, variance, data):
    """Mean log loss.

    :param mean: (tensor) mean of prediction.
    :param variance: (tensor) variance of prediction.
    :param data: (tensor) reference data.
    """
    return (0.5 * np.log(2 * np.pi * variance) +
            0.5 * (mean - data) ** 2 / variance).mean()


def smll(mean, variance, data):
    """Standardised mean log loss.

    :param mean: (tensor) mean of prediction.
    :param variance: (tensor) variance of prediction.
    :param data: (tensor) reference data.
    """
    return mll(mean, variance, data) - mll(data.mean(), data.var(ddof=0), data)


def r2(mean, data):
    """R-squared.

    :param mean: (tensor) mean of prediction.
    :param data: (tensor) reference data.
    """
    return 1 - ((data - mean) ** 2).mean() / data.var(ddof=0)



import glob
import multiprocessing
import os
import pickle
from itertools import product

import numpy as np
import pandas as pd

from .data import data_path, split_df, resource, dependency

__all__ = ['load']

cache_data = data_path('eeg', 'data.pickle')
cache_experiment = data_path('eeg', 'experiment.pickle')


def load():
    _fetch()

    # Generate cache if it does not exist.
    if not os.path.exists(cache_experiment):
        _parse()

    with open(cache_experiment, 'rb') as f:
        return pickle.load(f)


def _fetch():
    # Training data:
    resource(target=data_path('eeg', 'SMNI_CMI_TRAIN.tar.gz'),
             url='https://archive.ics.uci.edu/ml/'
                 'machine-learning-databases/eeg-mld/'
                 'SMNI_CMI_TRAIN.tar.gz')
    dependency(target=data_path('eeg', 'train'),
               source=data_path('eeg', 'SMNI_CMI_TRAIN.tar.gz'),
               commands=['tar xzf SMNI_CMI_TRAIN.tar.gz',
                         'mv SMNI_CMI_TRAIN train',
                         'find train | grep gz$ | xargs gunzip'])

    # Test data:
    resource(target=data_path('eeg', 'SMNI_CMI_TEST.tar.gz'),
             url='https://archive.ics.uci.edu/ml/'
                 'machine-learning-databases/eeg-mld/'
                 'SMNI_CMI_TEST.tar.gz')
    dependency(target=data_path('eeg', 'test'),
               source=data_path('eeg', 'SMNI_CMI_TEST.tar.gz'),
               commands=['tar xzf SMNI_CMI_TEST.tar.gz',
                         'mv SMNI_CMI_TEST test',
                         'find test | grep gz$ | xargs gunzip'])


def _parse_trial(fp):
    rows = np.genfromtxt(fp, delimiter=' ', dtype=str)
    sites = {}

    # Extract site names.
    site_names = tuple(sorted(set([x[1] for x in rows])))

    # Extract sites.
    for row in rows:
        try:
            sites[row[1]].append((row[2], row[3]))
        except KeyError:
            sites[row[1]] = [(row[2], row[3])]

    # Convert to data series, assuming that all inputs are the same for all
    # sites.
    x, ys = None, []
    for site in sorted(sites.keys()):
        rows = np.array(sites[site], dtype=float)
        x = rows[:, 0] / 256.  # Sampled at 256 Hz.
        ys.append(rows[:, 1])

    return {'df': pd.DataFrame(np.stack(ys, axis=0).T,
                               index=pd.Index(x, name='time'),
                               columns=site_names)}


def _extract_trials(fps):
    # Extract trial numbers.
    trial_numbers = map(lambda x: int(x.split('.rd.')[1]), fps)

    # Parse trials.
    with multiprocessing.Pool(processes=8) as pool:
        parsed_trials = pool.map(_parse_trial, fps)

    # Check that all lists of site names are the same.
    if len(set([tuple(d['df'].columns) for d in parsed_trials])) != 1:
        raise AssertionError('Site names are inconsistent between trials.')

    # Return dictionary mapping trial number to parsed result.
    return {k: v for k, v in zip(trial_numbers, parsed_trials)}


def _parse():
    print('Parsing EEG data. This may take a while.')
    numbers = [('c', n) for n in [337, 338, 339, 340, 341,
                                  342, 344, 345, 346, 347]] + \
              [('a', n) for n in [364, 365, 368, 369, 370,
                                  371, 372, 375, 377, 378]]
    partitions = ['train', 'test']
    subject_dir_format = data_path('eeg', '{partition}',
                                   'co2{type}{subject_n:07d}')

    # Create containers for the different partitions.
    data = {partition: {} for partition in partitions}

    # Iterate over all subjects.
    for partition, (subject_type, subject_n) in product(partitions, numbers):

        # Create an entry for the current subject.
        if subject_n not in data[partition]:
            data[partition][subject_n] = {'type': subject_type}

        # Determine directory of subject.
        subject_dir = subject_dir_format.format(partition=partition,
                                                type=subject_type,
                                                subject_n=subject_n)

        # Get all trials files and extract data.
        trial_files = glob.glob(subject_dir + '/*.rd.*')
        data[partition][subject_n]['trials'] = _extract_trials(trial_files)

        # Tag data series with subject number and subject type.
        for d in data[partition][subject_n]['trials'].values():
            d['label'] = (subject_n, subject_type)

    # Checks that all lists of site names are consistent.
    if len(set([tuple(d['df'].columns)
                for _, n in numbers
                for p in partitions
                for d in data[p][n]['trials'].values()])) != 1:
        raise AssertionError('Site names are inconsistent between subjects.')

    # Dump extracted data to file.
    with open(cache_data, 'wb') as f:
        pickle.dump(data, f)

    # For the experiment, just use a single trial.
    trial = data['train'][337]['trials'][0]

    # Select a number of fixed labels.
    labels = ['F3', 'F4', 'F5', 'F6'] + ['FZ', 'F1', 'F2']
    trial = trial['df'][labels]

    # Split according to paper.
    test, train = split_df(trial,
                           index_range=(trial.shape[0] - 100, trial.shape[0]),
                           columns=['FZ', 'F1', 'F2'],
                           iloc=True)

    # Save experiment data.
    with open(cache_experiment, 'wb') as f:
        pickle.dump((trial, train, test), f)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()


setuptools.setup(
        name='sgpvae',
        version='0.1',
        author='Matthew Ashman',
        author_email='mca39@cam.ac.uk',
        description='Sparse Gaussian process variational autoencoders',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/MattAshman/sgpvae',
        packages=setuptools.find_packages(),
        classifiers=[
                'Programming Language :: Python :: 3',
                'License :: OSI Approved :: MIT License',
                'Operating System :: OS Independent',
        ],
        python_requires='>3.6',
)




import torch
import torch.nn as nn

from .kernels import Kernel

__all__ = ['KernelList', 'AdditiveKernel', 'MultiplicativeKernel']


class KernelList(nn.ModuleList):
    """A list of kernels.

    :param kernels (list): list of kernels.
    """
    def __init__(self, kernels):
        super().__init__(kernels)

    def forward(self, x1, x2, diag=False, embed=True):
        """Returns the covariance matrix defined by kernels in list."""
        covs = [kernel.forward(x1, x2, diag) for kernel in self]

        if diag and embed:
            covs = torch.stack([cov.diag_embed() for cov in covs])
        else:
            covs = torch.stack(covs)

        return covs


class AdditiveKernel(Kernel):
    """The addition of kernels.

    :param args (list): list of kernels.
    """
    def __init__(self, *args):
        super().__init__()

        self.kernels = KernelList(args)

    def forward(self, x1, x2, diag=False):
        """Returns the covariance matrix defined by the addition of kernels
        in list."""
        cov = self.kernels.forward(x1, x2, diag, embed=False).sum(0)

        return cov


class MultiplicativeKernel(Kernel):
    """The product of kernels.

    :param args (list): list of Kernels.
    """
    def __init__(self, *args):
        super().__init__()

        self.kernels = KernelList(args)

    def forward(self, x1, x2, diag=False):
        """Returns the covariance matrix defined by the product of kernels
        in list."""
        cov = self.kernels.forward(x1, x2, diag, embed=False).prod(0)

        return cov



import pandas as pd

from .data import data_path, resource

__all__ = ['load']


def load():
    _fetch()

    name_corr = {'Xloc': 'x',
                 'Yloc': 'y',
                 'Landuse': 'land',
                 'Rock': 'rock'}

    # Load training data.
    columns = pd.read_csv(data_path('jura', 'prediction.dat'),
                          skiprows=2,
                          header=None,
                          nrows=11)
    columns = [name_corr[c] if c in name_corr else c for c in columns[0]]
    train = pd.read_csv(data_path('jura', 'prediction.dat'),
                        skiprows=13,
                        header=None,
                        sep=r'\s+')
    train.columns = columns
    train.set_index(['x', 'y'], inplace=True)

    # Load test data.
    test = pd.read_csv(data_path('jura', 'validation.dat'), sep='\t')
    test.columns = [name_corr[c] if c in name_corr else c for c in test.columns]
    test.set_index(['x', 'y'], inplace=True)

    # Setup according to experiment.
    train = pd.concat([train[['Ni', 'Zn', 'Cd']], test[['Ni', 'Zn']]])
    test = test[['Cd']]

    return train, test


def _fetch():
    resource(target=data_path('jura', 'prediction.dat'),
             url='https://drive.google.com/uc'
                 '?export=download&id=0B6subVejOkD_Vm8teTQtWjB3QzQ')
    resource(target=data_path('jura', 'validation.dat'),
             url='https://drive.google.com/uc'
                 '?export=download&id=0B6subVejOkD_UjFadE5qVHNRTGM')



import torch.nn as nn

from torch.nn import functional as F

__all__ = ['LinearNN']


class LinearNN(nn.Module):
    """A fully connected neural network.

    :param in_dim (int): dimension of the input variable.
    :param out_dim (int): dimension of the output variable.
    :param hidden_dims (list, optional): dimensions of hidden layers.
    :param nonlinearity (function, optional): non-linearity to apply in
    between layers.
    """
    def __init__(self, in_dim, out_dim, hidden_dims=(64, 64),
                 nonlinearity=F.relu):
        super().__init__()

        self.nonlinearity = nonlinearity

        self.layers = nn.ModuleList()
        for i in range(len(hidden_dims) + 1):
            if i == 0:
                self.layers.append(nn.Linear(in_dim, hidden_dims[i]))
            elif i == len(hidden_dims):
                self.layers.append(nn.Linear(hidden_dims[i-1], out_dim))
            else:
                self.layers.append(nn.Linear(hidden_dims[i-1], hidden_dims[i]))

        # Weight initialisation.
        for layer in self.layers:
            nn.init.xavier_normal_(layer.weight)
            if layer.bias.data is not None:
                nn.init.constant_(layer.bias, 0)

    def forward(self, x):
        """Returns output of the network."""
        for layer in self.layers[:-1]:
            x = self.nonlinearity(layer(x))

        x = self.layers[-1](x)
        return x



import copy
import torch
import numpy as np

__all__ = ['elbo_subset', 'elbo_mm']


def elbo_subset(model, x, y, mask, num_samples=1, k=1, p=0.5):
    elbo = 0
    for _ in range(k):
        mask_q = copy.deepcopy(mask)
        for n, row in enumerate(mask):
            row_q = copy.deepcopy(row)
            data_idx = torch.where(row == 1)[0]
            to_remove = np.random.binomial(len(data_idx) - 1, p=p)
            remove_idx = np.random.choice(data_idx.numpy(), size=[to_remove],
                                          replace=False)

            row_q[remove_idx] = 0
            mask_q[n, :] = row_q

        elbo += model.elbo(x, y, mask, mask_q, num_samples)

    return elbo


def elbo_mm(model, x, y, mask, num_samples, k, p=0.5):
    """Monte Carlo estimate of the evidence lower bound from the Multimodal
    VAE."""
    elbo = model.elbo(x, y, mask, num_samples=num_samples)

    elbos_n = 0
    for n in range(y.shape[1]):
        mask_q = copy.deepcopy(mask)
        mask_q[:, n] = 0
        elbos_n += model.elbo(x, y, mask, mask_q, num_samples)

    elbos_k = 0
    for _ in range(k):
        mask_q = copy.deepcopy(mask)
        for n, row in enumerate(mask):
            row_q = copy.deepcopy(row)
            data_idx = torch.where(row == 1)[0]
            to_remove = np.random.binomial(len(data_idx)-1, p=p)
            remove_idx = np.random.choice(data_idx.numpy(), size=[to_remove],
                                          replace=False)

            row_q[remove_idx] = 0
            mask_q[n, :] = row_q

        elbos_k += model.elbo(x, y, mask, mask_q, num_samples)

    elbo += elbos_n + elbos_k

    return elbo
